/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: simplest.c,v 1.111 2010/07/01 05:50:45 zlb Exp $
 *
 * This sample code solves the Helmholtz equation:
 * 	 -\Delta u + a u = f
 * with Dirichlet or periodic boundary conditions */

#include "phg.h"
#include <string.h>
#include <math.h>

static FLOAT a = 1.0;
static FLOAT b = 1.0;

#define TEST_CASE 5

static void
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
#if TEST_CASE == 0
    *value = Cos(2. * M_PI * x) * Cos(2. * M_PI * y) * Cos(2. * M_PI * z);
#elif TEST_CASE == 1
    *value = x + 1.2*y - 0.4 * z;
#elif TEST_CASE == 2
#  warning test case 2
    *value = x*x;
#elif TEST_CASE == 3
#  warning test case 3
    if (x > .5)
	*value = 1;
    else
	*value = -1;
#elif TEST_CASE == 4
#  warning test case 4
    if (z > .5)
	*value = x + 0.5*y - 0.7*z;
    else
	*value = x + 0.5*y - 0.7*z + 1;
#elif TEST_CASE == 5
#  warning test case 5
    if (z > .5)
	*value = y;
    else
	*value = x;
#endif
}

static void
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    func_u(x, y, z, value);
#if TEST_CASE == 0
    *value = b * 12. * M_PI * M_PI * *value
	+ a * *value;
#else
    *value = b * 0.
	+ a * *value;
#endif
}

static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int i, j;
    GRID *g = u_h->g;
    SIMPLEX *e;

    assert(u_h->dim == 1);
    ForAllElements(g, e) {
	int q, N = DofGetNBas(u_h, e);	/* number of basises in the element */
	FLOAT A[N][N], rhs[N], buffer[N];
	INT I[N];
	QUAD *quad;
	const FLOAT *vf, *w, *p;
	FLOAT vol;

	for (i = 0; i < N; i++) 
	    I[i] = phgSolverMapE2L(solver, 0, e, i);

	vol = phgGeomGetVolume(g, e);
	quad = phgQuadGetQuad3D(4);
	vf = phgQuadGetDofValues(e, f_h, quad);
	p = quad->points;
	w = quad->weights;

	bzero(A, sizeof(A));
	bzero(rhs, sizeof(rhs));

	for (q = 0; q < quad->npoints; q++) {
	    for (i = 0; i < N; i++) {
		const FLOAT *gi = phgQuadGetBasisValues(e, u_h, i, quad) + q;
		for (j = 0; j < N; j++) {
		    const FLOAT *gj = phgQuadGetBasisValues(e, u_h, j, quad) + q;
		    A[i][j] += vol * (*w) * (
					     a * (*gi) * (*gj)
					     );
		}
		rhs[i] += vol*(*w) * (*vf) * (*gi);
	    }
	    vf++;
	    p += Dim + 1;
	    w++;
	}

	/* FLOAT vol; */
	/* printf(" nbas: %d\n", N); */
 	/* phgPrintf("Element matrix: %e\n", vol); */
	/* for (i = 0; i < N; i++) { */
	/*     for (j = 0; j < N; j++) { */
	/* 	phgPrintf("%15.4e, ", A[i][j] / vol); */
	/*     } */
	/*     phgPrintf("\n"); */
	/* } */

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, func_u, buffer, rhs+i,
					DOF_PROJ_NONE)) {
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, buffer); 
	    }
	    else {	/* interior node */
		/* right hand side = \int f * phi_i */
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, A[i]); 
	    }
	}
	phgSolverAddRHSEntries(solver, N, I, rhs);
    }

    phgMatDumpMATLAB(solver->mat, "A", "A_.m");
}

static void
estimate_error(DOF *u_h, DOF *f_h, DOF *grad_u, DOF *error)
/* compute H1 error indicator */
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *jump, *residual, *tmp;

    jump = phgQuadFaceJump(grad_u, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    tmp = phgDofDivergence(grad_u, NULL, NULL, NULL);
    residual = phgDofGetSameOrderDG(u_h, -1, NULL);
    phgDofCopy(f_h, &residual, NULL, NULL); 
    phgDofAXPY(-a, u_h, &residual);
    phgDofAXPY(1.0, tmp, &residual);
    phgDofFree(&tmp);
    ForAllElements(g, e) {
	int i;
	FLOAT eta, h;
	FLOAT diam = phgGeomGetDiameter(g, e);
	eta = 0.0;
	/* for each face F compute [grad_u \cdot n] */
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & (DIRICHLET | NEUMANN))
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    eta += *DofFaceData(jump, e->faces[i]) * h;
	}
	eta = eta * .5 + diam * diam * phgQuadDofDotDof(e, residual, residual,
								QUAD_DEFAULT);
	*DofElementData(error, e->index) = eta;
    }
    phgDofFree(&jump);
    phgDofFree(&residual);

    return;
}

int
main(int argc, char *argv[])
{
    INT periodicity = 0 /* X_MASK | Y_MASK | Z_MASK */;
    INT mem_max = 400;
    FLOAT tol = 5e-1;
    char *fn = "../test/cube4.dat";
    GRID *g;
    SIMPLEX *e;
    DOF *u_h, *f_h, *grad_u, *error, *u;
    SOLVER *solver;
    FLOAT L2error, indicator;
    BOOLEAN use_hp = FALSE;
	int pre_refines = 0;
    size_t mem, mem_peak;

    phgOptionsRegisterNoArg("use_hp", "Use HP", &use_hp);
    phgOptionsRegisterFloat("a", "Coefficient", &a);
    phgOptionsRegisterFloat("b", "Coefficient", &b);
    phgOptionsRegisterFloat("tol", "Tolerance", &tol);
    phgOptionsRegisterInt("pre_refines", "Pre refines", &pre_refines);
    phgOptionsRegisterInt("mem_max", "Maximum memory (MB)", &mem_max);
    phgOptionsRegisterInt("periodicity", "Set periodicity", &periodicity);

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    phgSetPeriodicity(g, periodicity);

    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    phgRefineAllElements(g, pre_refines);

    /* The discrete solution */
    if (use_hp) {
	/* u_h is h-p type */
	HP_TYPE *hp = phgHPNew(g, HP_HB);
#if 0
	/* random */
	ForAllElements(g, e)
	    e->hp_order = 1 + GlobalElement(g, e->index) % 2;
#else
	/* conditional */
	ForAllElements(g, e) {
	    FLOAT lambda[4] = {.25, .25, .25, .25};
	    FLOAT x, y, z;
	    phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
	    if (z > .5)
		e->hp_order = 2;
	    else
		e->hp_order = 1;
	}
#endif
	phgHPSetup(hp, FALSE);
	phgInfo(0, "HP %s %s %s\n", 
	       hp->info->types[0]->name,
	       hp->info->types[1]->name,
	       hp->info->types[2]->name
	       );
	u_h = phgHPDofNew(g, hp, 1, "u_h", DofInterpolation);
	phgHPFree(&hp);
    }
    else {
	/* u_h is non h-p type */
	u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    }
    phgDofSetDataByValue(u_h, 0.0);
    u_h->DB_mask = 0;

    /* RHS function */
    f_h = phgDofNew(g, DOF_ANALYTIC, 1, "f_h",  func_f);

    /* DOF for storing a posteriori error estimates */
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);

    /* The analytic solution */
    u = phgDofNew(g, DOF_ANALYTIC, 1, "u", func_u);
    if (!DofIsHP(u_h)) {
	phgPrintf("#   Dof type: %s\n", u_h->type->name);
    }

    int STEP = 1;
    while (TRUE) {
        if (STEP > 10) break;
        phgPrintf("STEP = %d\n", STEP);
        STEP++;
	double t0 = phgGetTime(NULL);
	if (phgBalanceGrid(g, 1.2, 1, NULL, 0.))
	    phgPrintf("Repartition mesh, load imbalance: %lg\n",
			(double)g->lif);
	phgPrintf("%d DOF, %d elements, %d submeshes, load imbalance: %lg\n",
			DofGetDataCountGlobal(u_h), g->nleaf_global, g->nprocs,
			(double)g->lif);
	solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
	phgPrintf("  DOF: %d, unknowns: %d, Dirichlet bdry: %d\n",
		DofGetDataCountGlobal(u_h), solver->rhs->map->nglobal,
		solver->rhs->map->bdry_nglobal);
	build_linear_system(solver, u_h, f_h);
	phgSolverSolve(solver, TRUE, u_h, NULL);
	phgPrintf("  nits = %d, ", solver->nits);
	phgSolverDestroy(&solver);
	grad_u = phgDofGradient(u_h, NULL, NULL, NULL);
	estimate_error(u_h, f_h, grad_u, error);
	indicator = Sqrt(phgDofNormL1Vec(error));
	phgDofFree(&grad_u);
	grad_u = phgDofCopy(u_h, NULL, NULL, NULL);
	phgDofAXPY(-1.0, u, &grad_u);
	L2error = phgDofNormL2(grad_u);
	phgDofFree(&grad_u);
	mem = phgMemoryUsage(g, &mem_peak);
	phgPrintf("L2 error = %0.3le, indicator = %0.3le, mem = %0.2lfMB\n",
			(double)L2error, (double)indicator,
			(double)mem_peak / (1024.0 * 1024.0));
	phgPrintf("  Wall time: %0.3le\n", phgGetTime(NULL) - t0);

	break;
	phgMarkRefine(MARK_DEFAULT, error, Pow(0.8,2), NULL, 0., 1,
			Pow(tol, 2) / g->nleaf_global);
	phgRefineMarkedElements(g);
    }

//    phgDofDump(u_h);

    FLOAT err = 0;
    ForAllElements(g, e) {
	QUAD *quad;
	const FLOAT *vu1, *vu0, *w, *p;
	int q;
	FLOAT vol;

	vol = phgGeomGetVolume(g, e);
 
	quad = phgQuadGetQuad3D(6);
	vu0 = phgQuadGetDofValues(e, u_h, quad);
	vu1 = phgQuadGetDofValues(e, u, quad);
	p = quad->points;
	w = quad->weights;

	for (q = 0; q < quad->npoints; q++) {
	    
	    err += vol*(*w) * (*vu0 - *vu1) * (*vu0 - *vu1);

	    vu0 ++;
	    vu1 ++;
	    p += Dim+1;
	    w++;
	}
	
    }

    phgPrintf("\n  Error L2: %e\n\n", err);
    DOF *u0 = phgDofCopy(u_h, NULL, DOF_DG0, "u_P0");

#if 0
    phgPrintf("Final mesh written to \"%s\".\n",
	phgExportDX(g, "simplest.dx", u_h, error, NULL));
#elif 1
    phgPrintf("Final mesh written to \"%s\".\n",
	      phgExportVTK(g, "simplest.vtk", u0, u_h, error, NULL));
#endif

    phgDofFree(&u);
    phgDofFree(&u0);
    phgDofFree(&u_h);
    phgDofFree(&f_h);
    phgDofFree(&error);

    phgFreeGrid(&g);

    phgFinalize();

    return 0;
}
