/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: lagrange.c,v 1.37 2012/01/07 01:08:58 zlb Exp $ */

/* Lagrange finite elements: DOF_G0, DOF_G1, DOF_G2, DOF_G3, DOF_G4, ...
 *
 * Number of terms of Pn = (n+1)(n+2)(n+3) / 3!
 */

#include "phg.h"

#include <string.h>


/*--------------------------- G1 element ----------------------------*/

static const FLOAT *
G1_bas(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
/* evaluation of basis functions */
{
    static FLOAT *buffer = NULL;
    FLOAT *p, *values;
    FLOAT L0 = lambda[0], L1 = lambda[1], L2 = lambda[2], L3 = lambda[3];

    Unused(dof);
    Unused(e);

    if (buffer == NULL) {
#if USE_OMP
#pragma omp critical (G1_bas)
#endif	/* USE_OMP */
	if (buffer == NULL) {
	    buffer = phgAlloc(phgMaxThreads * (NVert + 1) * sizeof(*buffer));
	    FreeAtExitNoCheck(buffer);
	}
    }

    p = values = buffer + phgThreadId * (NVert + 1);

    if (no1 <= 0)
	no1 = NVert + 1;
    assert(no0 < no1);

    if (no0 == 0 && no1 == NVert + 1) {
	*(p++) = L0;
	*(p++) = L1;
	*(p++) = L2;
	*(p++) = L3;
	*p     = 1;
	return values;
    }

    switch (no0) {
    case 0:
	*(p++) = L0;
	if (no1 == 1)
	    return values;
    case 1:
	*(p++) = L1;
	if (no1 == 2)
	    return values;
    case 2:
	*(p++) = L2;
	if (no1 == 3)
	    return values;
    case 3:
	*(p++) = L3;
	if (no1 == 4)
	    return values;
    case 4:
	*(p++) = 1.;
	if (no1 == 5)
	    return values;
    }

    return values;
}

static const FLOAT *
G1_grad(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
/* evaluation of gradients of basis functions */
{
    static FLOAT (*buffer)[Dim + 1] = NULL;
    FLOAT (*values)[Dim + 1];
    FLOAT L0 = lambda[0], L1 = lambda[1], L2 = lambda[2], L3 = lambda[3];
    int i, j;
 
    Unused(dof);
    Unused(e);

    if (buffer == NULL) {
#if USE_OMP
#pragma omp critical (G1_bas)
#endif	/* USE_OMP */
	if (buffer == NULL) {
	    buffer = phgAlloc(phgMaxThreads * (NVert + 1) * sizeof(*buffer));
	    FreeAtExitNoCheck(buffer);
	}
    }

    values = buffer + phgThreadId * (NVert + 1);

    if (no1 <= 0)
	no1 = NVert + 1;
    assert(no0 < no1);

    for (j = no0; j < no1; j++)
	if (j < NVert)
	    for (i = 0; i < Dim + 1; i++)
		values[j][i] = (i == j) ? 1.0 : 0.0;

    return (FLOAT *)(values + no0);
}

static void
G1_interp(DOF *dof, SIMPLEX *e, FLOAT **old_data, FLOAT **new_data)
/* interpolation */
{
    int i;
    FLOAT *vnew = new_data[0], *vold0 = old_data[0], *vold1 = old_data[1];

    Unused(e);

    phgError(1, "G1 interp unimplemented.\n");
    for (i = 0; i < dof->dim; i++)
	*(vnew++) = 0.5 * (*(vold0++) + *(vold1++));
}



/*-------------------------------------------------------------------*/


static FLOAT G1_points[] = {_F(1.0), _F(0.25), _F(0.25), _F(0.25), _F(0.25)};
DOF_TYPE DOF_G1_ = {DofReserved, "G1", G1_points, NULL, DOF_DG0, NULL, NULL,
		    G1_interp, phgDofInterpF2CGeneric, phgDofInitFuncPoint,
		    G1_bas, G1_grad, NULL, FE_L2,
		    TRUE, FALSE, -1,    /* invariant, free_after_use, id */
		    NVert + 1, 1, -1, 1, /* nbas, order, cont, dim */
		    1, 0, 0, 1};        /* np_vert, np_edge, np_face, np_elem */


DOF_TYPE *DOF_Gn[] = {DOF_P0, DOF_G1, NULL};
